public class Channel {
	
	// ==================================================================================
	// globals / pre-constructor
	// ==================================================================================

	// settings
	0 => int id;
	16 => int length;
	
	// state
	float mute;
	float solo;
	float volume;

	int rootNote;
	int mainPitch;
	int thisPitch;
	
	float stepVolumes[length];
	int stepPitches[length];
	
	// sound buffer and channel data
	sndbuf soundBuffer;
	gain channelGain;


	// ==================================================================================
	// init
	// ==================================================================================
	
	fun void init(int i, int l) {
		i => id;
		setLength(l);
		
		0 => mute;
		0 => solo;
		0.5 => volume;
		60 => rootNote;
	}
	
	fun void setLength(int l) {
		l => length;
		
		float sv[l] @=> stepVolumes;
		int sp[l] @=> stepPitches;
		
		initPitches();
	}
	
	fun void initPitches() {
		for( 0 => int j; j < stepPitches.cap(); j++ ) {
			60 => stepPitches[j];
		}
	}
	
	// ==================================================================================
	// play
	// ==================================================================================
	
	fun void play(int step) {
		if((stepVolumes[step] == 0) || (mute == 1)) return; 
		
		// reset sample position
		0 => soundBuffer.pos;
			
		// calc pitch and set
		mainPitch + stepPitches[step] => thisPitch;
		if(thisPitch < 0) 0 => thisPitch;
		setSampleRate(thisPitch);
			
		// calc volume and play
		stepVolumes[step] * volume => soundBuffer.gain;
	}
	
	// set given sample buffer to play at given midi pitch
	fun void setSampleRate(int note) {
		rootNote => int root;
		std.mtof(note) / std.mtof(root) => soundBuffer.rate;
	}
	
	
	// ==================================================================================
	// util
	// ==================================================================================
	
	fun void debug() {
		for( 0 => int i; i < stepVolumes.cap(); i++ ) {
			<<< i , " " , stepVolumes[i] >>>;
		}
	}
}