// ==================================================================================
// Chuck step sequencer v3 (for Breakage v2)
// ==================================================================================

	// ==================================================================================
	// osc 
	// ==================================================================================
	
	3000 => int oscSendPort;
	2000 => int oscRecvPort;
	
	// create OSC sender
	OscSend oscXmit;
	oscXmit.setHost( "localhost", oscSendPort);
	
	// create OSC receiver
	OscRecv oscRecv;
	oscRecvPort => oscRecv.port;
	oscRecv.listen();
	
	// steps
	oscRecv.event( "/setStepVelocity,i,i,f" ) @=> OscEvent oscStepVelocityEvent;
	oscRecv.event( "/setStepPitch,i,i,i" ) @=> OscEvent oscStepPitchEvent;
	
	// sequencer
	oscRecv.event( "/setBpm,f" ) @=> OscEvent oscBpmEvent;
	oscRecv.event( "/setMasterVolume,f" ) @=> OscEvent oscMasterVolumeEvent;
	oscRecv.event( "/setSwing,f" ) @=> OscEvent oscSwingEvent;
	oscRecv.event( "/setPitch,i" ) @=> OscEvent oscPitchEvent;
	oscRecv.event( "/play,i" ) @=> OscEvent oscPlayEvent;

	oscRecv.event( "/record,i" ) @=> OscEvent oscRecordingEvent;
	oscRecv.event( "/mainSync,i,i" ) @=> OscEvent oscMainSyncEvent;
	
	// channels
	oscRecv.event( "/setChannelMute,i,i" ) @=> OscEvent oscChannelMuteEvent;
	oscRecv.event( "/setChannelSolo,i,i" ) @=> OscEvent oscChannelSoloEvent;
	oscRecv.event( "/setChannelVolume,i,f" ) @=> OscEvent oscChannelVolumeEvent;
	oscRecv.event( "/setChannelSample,i,s" ) @=> OscEvent oscChannelSampleEvent;
	
	// pattern
	oscRecv.event( "/setPatternDimensions,i,i" ) @=> OscEvent oscPatternDimensionEvent;
	
	// ==================================================================================
	// sequencer 
	// ==================================================================================
	
	// sequencer's step, cycle and state
	0 => int step;
	0 => int length;
	0 => int cycle;
	1 => int playState;		// set this to set the desired play state - 0 = play, 1 = stop, 2 = pause
	0 => int isPlaying;		// set by the main loop to give the actual play state - 0 not playing, 1 is playing
		
	// prepare timer 
	200 => float bpm;
	60 / bpm => float bps;
	bps::second => dur timer;
	timer - (now % timer) => now;
	
	// prepare swing
	0.0 => float swingAmount;
	bps => float thisBps;
	
	1 => int swing;
	
	// pitch
	0 => int pitch;
	
	// ==================================================================================
	// midi clock sync
	// ==================================================================================
	
	MidiIn min;
	MidiMsg msg;
	
	fun void openMidiInDevice(int device) {
		if( !min.open( device ) ) {
			<<< "Couldn't open MIDI in device" >>>;
		}
	}
	
	// is this still used?
	0 => int count;
	((24 * 4) / 16) => int countLimit;
	
	// ==================================================================================
	// channels volumes array
	// ==================================================================================
	
	gain masterGain;
		
	1 => int initializing;
	Channel channels[];
	
	// ================================
	// handshake with Breakage
	// ================================
	
	// send hello to Breakage
	<<< "hello Breakage, this is ChucK" >>>;
	oscXmit.startMsg( "/hello", "i" );
	1 => oscXmit.addInt;	
	
	// wait for init message from Breakage
	oscPatternDimensionEvent => now;
	<<< "ChucK received handshake" >>>;
		
	// grab the message
	while ( oscPatternDimensionEvent.nextMsg() != 0 ) { 
		oscPatternDimensionEvent.getInt() => int i;
		oscPatternDimensionEvent.getInt() => int j;		
		initDimensions(i, j);
	}
		
	fun void initDimensions(int num, int l) {
		// stop sequencer loop
		1 => playState;
		0 => step;
		
		// wait for the sequencer to actually pause
		while(isPlaying != 0) {
			0.05::second => now;
		}

		// stop osc threads		
		if(initializing != 1) {
			oscRecv.stop();
			0 => isListening;
			me.yield();
			2.25::second => now;
		}
		
		// sequence size
		l => length;
		Channel ch[num] @=> channels;
		
		for( 0 => int i; i < channels.cap(); i++ ) {
			channels[i].init(i, length);
			pitch => channels[i].mainPitch;
			channels[i].soundBuffer => masterGain;
		}	
		
		// restart if necessary
		if(step > channels[0].length) 0 => step;
		
		// restart sequencer loop
		0 => playState;
		if(initializing != 1) {
			oscRecv.listen();
			1 => isListening; 
		}
	}
	
	// ==================================================================================
	// samples: sound buffers
	// ==================================================================================
	
	// load default samples
	"projects/samples/kick.wav" => channels[0].soundBuffer.read;
	"projects/samples/snare.wav" =>  channels[1].soundBuffer.read;
	"projects/samples/closedhat.wav" =>  channels[2].soundBuffer.read;
	"projects/samples/openhat.wav" =>  channels[3].soundBuffer.read;

	// apply filter to channel 2 (closed hat)
	// soundBuffers[2] => biquad f => masterGain;
	
	// global fx
	PRCRev reverb;
	Chorus chorus;
	biquad f;
	
	//masterGain => chorus => reverb => dac;
	masterGain => dac;
	
	0.8 => masterGain.gain;
	
	0.15 => reverb.mix;
	
	0.15 => chorus.modFreq;
	0.15 => chorus.modDepth;
	0.45 => chorus.mix;
	
	// ==================================================================================
	// samples: filter
	// ==================================================================================
	
	.99 => f.prad;		// set biquad pole radius
	.05 => f.gain;		// set biquad gain
	1 => f.eqzs;		// set equal zeros 
	0.0 => float t;		// our float
	
	// ==================================================================================
	// inst: sitar
	// ==================================================================================
	
	// instrument - NOT USED
	Sitar sit => PRCRev r => dac;
	.15 => r.mix;
	
	// play function
	fun void playSitar(float f) {
	    f => sit.freq;
	
	    // pluck!
	    std.rand2f( 0.4, 0.9 ) => sit.noteOn;
	}

	// ==================================================================================
	// osc listener loops -- listening state
	// ==================================================================================
	1 => int isListening;
	
	// ==================================================================================
	// osc listener loops -- steps
	// ==================================================================================
	
	fun void stepPitchListener() {
		while ( isListening ) {
		    oscStepPitchEvent => now;  // wait for event to arrive
		
		    while ( oscStepPitchEvent.nextMsg() != 0 ) { 
		        oscStepPitchEvent.getInt() => int a;
		        oscStepPitchEvent.getInt() => int b;
		        oscStepPitchEvent.getInt() => int c;
		        c => channels[a].stepPitches[b];
		    }
		}
	}
	
	fun void stepVelocityListener() {
		while ( isListening ) {
		    oscStepVelocityEvent => now;  // wait for event to arrive
		
		    while ( oscStepVelocityEvent.nextMsg() != 0 ) { 
		        oscStepVelocityEvent.getInt() => int a;
		        oscStepVelocityEvent.getInt() => int b;
		        oscStepVelocityEvent.getFloat() => float c;
		        c => channels[a].stepVolumes[b];
		    }
		}
	}
	
	// ==================================================================================
	// osc listener loops -- sequencer
	// ==================================================================================
	
	// play state
	fun void startPlayListener() {
		while ( isListening ) {
		    oscPlayEvent => now;
		
		    // grab the next message from the queue. 
		    while ( oscPlayEvent.nextMsg() != 0 ) { 
				oscPlayEvent.getInt() => int i;
				
				// 0 = play, 1 = stop, 2 = pause
				
				if(i == 1) {
					// if we're stopping, reset the loop position
					1 => playState;
					length-1 => step;
					
				} else if(i == 2) {
					// if we're pausing, check state
					if(playState == 0) {
						2 => playState;
					} else {
						0 => playState;
					}
				
				} else {
					// playing, so pause or stop
					i => playState;
				}
		    }
		}
	}
	
	// master volume
	fun void masterVolumeListener() {
		while ( isListening ) {
		    oscMasterVolumeEvent => now;  // wait for event to arrive
		
		    while ( oscMasterVolumeEvent.nextMsg() != 0 ) { 
		        oscMasterVolumeEvent.getFloat() => float b;
		        b => masterGain.gain;
	///	        <<<"<set masterVolume=\"", b, "\" />" >>>;
		    }
		}
	}
	
	// bpm
	fun void bpmListener() {
		while ( isListening ) {
		    oscBpmEvent => now; // wait for event to arrive
		
		    while ( oscBpmEvent.nextMsg() != 0 ) { 
		        oscBpmEvent.getFloat() => bpm;
			60 / bpm => bps;
			bps::second => dur timer;
			timer - (now % timer) => now;
	//		<<<"<set bpm=", bpm, " />" >>>;
		    }
		}
	}
	
	// swing
	fun void swingListener() {
		while ( isListening ) {
		    oscSwingEvent => now; // wait for event to arrive
		
		    while ( oscSwingEvent.nextMsg() != 0 ) { 
		        oscSwingEvent.getFloat() => swingAmount;
	///			<<<"<set swing=", swingAmount, " />" >>>;
		    }
		}
	}
	
	// pitch
	fun void pitchListener() {
		while ( isListening ) {
		    oscPitchEvent => now; // wait for event to arrive
		
		    while ( oscPitchEvent.nextMsg() != 0 ) { 
		        oscPitchEvent.getInt() - 30 => pitch;
			
			for( 0 => int i; i < channels.cap(); i++ ) {
				pitch => channels[i].mainPitch;
			}

	//		<<<"<set pitch=", pitch, " />" >>>;
		    }
		}
	}
	
	-1 => int recShred;
	
	// recording
	fun void recordingListener() {
		while ( isListening ) {
		    oscRecordingEvent => now; // wait for event to arrive
		
		    while ( oscRecordingEvent.nextMsg() != 0 ) { 
		        oscRecordingEvent.getInt() => int r;
		        
		        <<< "record ", r >>>;
			
			if(r == 0) {			// stop recording
				if(recShred > -1) { 
					machine.add( "stereo-rec-stop.ck" ); // => recShred;
				}

			} else if (r == 1) {		// start
				machine.add( "stereo-rec.ck" ) => recShred;
				<<< "started" >>>;
			}
				
		    }
		}
	}
	
	// ==================================================================================
	// osc listener loops -- channels
	// ==================================================================================
	
	// mute
	fun void channelMuteListener() {
		while ( isListening ) {
		    oscChannelMuteEvent => now;  // wait for event to arrive
		
		    while ( oscChannelMuteEvent.nextMsg() != 0 ) { 
		        oscChannelMuteEvent.getInt() => int a;
		        oscChannelMuteEvent.getInt() => int b;
		        b => channels[a].mute;
		        <<<"<set channel=\"", a, "\" mute=\"",b,"\" />" >>>;
		    }
		}
	}
	
	// solo
	fun void channelSoloListener() {
		while ( isListening ) {
		    oscChannelSoloEvent => now;  // wait for event to arrive
		
		    while ( oscChannelSoloEvent.nextMsg() != 0 ) { 
		        oscChannelSoloEvent.getInt() => int a;
		        oscChannelSoloEvent.getInt() => int b;
		        b => channels[a].solo;
	///	        <<<"<set channel=\"", a, "\" solo=\"",b,"\" />" >>>;
		    }
		}
	}
	
	// channel volume
	fun void channelVolumeListener() {
		while ( isListening ) {
		    oscChannelVolumeEvent => now;  // wait for event to arrive
		
		    while ( oscChannelVolumeEvent.nextMsg() != 0 ) { 
		        oscChannelVolumeEvent.getInt() => int a;
		        oscChannelVolumeEvent.getFloat() => float b;
		        b => channels[a].volume;
	///	        <<<"<set channel=\"", a, "\" volume=\"",b,"\" />" >>>;
		    }
		}
	}
	
	// open sample
	fun void setChannelSampleListener() {
		while ( isListening ) {
		    oscChannelSampleEvent => now;
		
		    // grab the next message from the queue. 
		    while ( oscChannelSampleEvent.nextMsg() != 0 ) { 
		        oscChannelSampleEvent.getInt() => int i;
		        oscChannelSampleEvent.getString() => string s;
		        s => channels[i].soundBuffer.read;
	//	        <<<"<set channel=\"", i, "\" sample=\"",s,"\" />" >>>;
		    }
		}
	}
	
	// ==================================================================================
	// osc listener loops -- pattern
	// ==================================================================================
	
	// pattern dimension
	fun void startPatternDimensionListener() {
		while ( isListening ) {
		    oscPatternDimensionEvent => now;
		
		    // grab the next message from the queue. 
		    while ( oscPatternDimensionEvent.nextMsg() != 0 ) { 
				oscPatternDimensionEvent.getInt() => int i;
				oscPatternDimensionEvent.getInt() => int j;
				
				// <<< "oscPatternDimensionEvent ", i, " x ", j >>>;
				initDimensions(i, j);
		    }
		}
	}
	
	// =============================
	// filter loop
	// =============================
	
	fun void filter() {
		//while( true ) {
	    	// sweep the filter resonant frequency
		    //100.0 + std.abs(math.sin(t)) * 1000.0 => f.pfreq;
	    	//t + .01 => t;
		    // advance time
	    	//100::ms => now;
		//}
	}
	//	spork ~ filter();

	// ==================================================================================
	// sporkage
	// ==================================================================================
	
	Shred patDimS;

	Shred chanSampS;
	Shred chanVolS;
	Shred chanSoloS;
	Shred chanMuteS;

	Shred stepPitchS;
	Shred stepVelS;

	Shred bpmS;
	Shred swingS;
	Shred pitchS;
	Shred masterVolS;
	Shred playS;
	Shred recordS;

	fun void sporkAll() {
		spork ~ startPatternDimensionListener() @=> patDimS;

		spork ~ setChannelSampleListener() @=> chanSampS;
		spork ~ channelVolumeListener() @=> chanVolS;
		spork ~ channelSoloListener() @=> chanSoloS;
		spork ~ channelMuteListener() @=> chanMuteS;

		spork ~ stepPitchListener() @=> stepPitchS;
		spork ~ stepVelocityListener() @=> stepVelS;
		
		spork ~ bpmListener() @=> bpmS;
		spork ~ swingListener() @=> swingS;
		spork ~ pitchListener() @=> pitchS;
		spork ~ masterVolumeListener() @=> masterVolS;
		spork ~ startPlayListener() @=> playS;
		spork ~ recordingListener() @=> recordS;

		me.yield();
	}

	// not used
	fun void killAllShreds() {
		patDimS.exit();
		chanSampS.exit();
		chanVolS.exit();
		chanSoloS.exit();
		chanMuteS.exit();

		stepPitchS.exit();
		stepVelS.exit();
		
		bpmS.exit();
		swingS.exit();
		pitchS.exit();
		masterVolS.exit();
		playS.exit();
		recordS.exit();

		250::ms => now;
		me.yield();
	}

	sporkAll();
	0 => initializing;

	// =======================================================================
	// step seq loop
	// =======================================================================
	
	//<<< "<breakage channels=\"", channels, "\" length=\"", channelLength, "\" />" >>>;
	//<<< "<set bpm=", bpm, " />" >>>;
		
	// ==========================================
	// main loop - internal timer
	// ==========================================
	
	fun void main() {
		while( true ) {
					
			// are we paused? 
			while(playState != 0) {
				0 => isPlaying;
				0.05::second => now;
			}
			1 => isPlaying;		
			
			// print step & cycle
		//	<<< "</step>" >>>;
		//	<<< "<step step=\"", step ,"\" cycle=\"", cycle, "\" >" >>>;
			
			// send step message
			oscXmit.startMsg( "/currentStep", "i" );
			step => oscXmit.addInt;
		
			// any solos?
			0 => int solos;
		
			for( 0 => int i; i < channels.cap(); i++) {
				if(channels[i].solo != 0) {
					1 => solos;
				}
			}
			
			// play sound
			for( 0 => int i; i < channels.cap(); i++ ) {
				// if there are solos, is this channel soloing?
				if((solos == 1) && (channels[i].solo == 1)) {
					channels[i].play(step);
			    
				// else should this channel play?
				} else if(solos == 0) {
					channels[i].play(step);
					// <<< "<play channel=\"", i,"\" volume=\"", stepVolumes[i][step], "\" pitch=\"" , stepPitches[i][step], "\" />" >>>;
				}
			}
			
			// advance time & swing
			if(swing == 1) {
				0 => swing;
				bps + swingAmount => thisBps;
			} else {
				1 => swing;
				bps - swingAmount => thisBps;
			}
		    
			thisBps::timer => now;
		        
			// increase step
			1 +=> step;
		
			// loop sequence if we've reached the end
			if( step == channels[0].length ) {
				0 => step;
				1 +=> cycle;
			}
		}
	}
	
	// ==========================================
	// main loop - MIDI in
	// ==========================================
	
	fun void mainSync() {
		while( true ) {
			
			// are we paused? 
			while(playState != 0) {
				0 => isPlaying;
				0.05::second => now;
			}
			1 => isPlaying;	
			
			
			// wait on event
			min => now;
    
  			// get the midi clock messages
    			min.recv(msg);
			
			// print step & cycle
		//	<<< "</step>" >>>;
		//	<<< "<step step=\"", step ,"\" cycle=\"", cycle, "\" >" >>>;
			
			// send step message
			oscXmit.startMsg( "/currentStep", "i" );
			step => oscXmit.addInt;
		
			// any solos?
			0 => int solos;
		
			for( 0 => int i; i < channels.cap(); i++) {
				if(channels[i].solo != 0) {
					1 => solos;
				}
			}
			
			// play sound
			for( 0 => int i; i < channels.cap(); i++ ) {
				// if there are solos, is this channel soloing?
				if((solos == 1) && (channels[i].solo == 1)) {
					channels[i].play(step);
			    
				// else should this channel play?
				} else if(solos == 0) {
					channels[i].play(step);
					// <<< "<play channel=\"", i,"\" volume=\"", stepVolumes[i][step], "\" pitch=\"" , stepPitches[i][step], "\" />" >>>;
				}
			}
			
			// advance time & swing
			if(swing == 1) {
				0 => swing;
				bps + swingAmount => thisBps;
			} else {
				1 => swing;
				bps - swingAmount => thisBps;
			}
		    
			thisBps::timer => now;
		        
			// increase step
			1 +=> step;
		
			// loop sequence if we've reached the end
			if( step == channels[0].length ) {
				0 => step;
				1 +=> cycle;
			}
		}
	}
	
	// ==========================================
	// master controller - daemonic
	// ==========================================
	
	Shred mainS;
	spork ~ main() @=> mainS;

	// pattern dimension
	fun void oscMainSyncListener() {
		while ( true ) {
			oscMainSyncEvent => now;
			
			// grab the next message from the queue. 
			while ( oscMainSyncEvent.nextMsg() != 0 ) { 
				oscMainSyncEvent.getInt() => int i;
				oscMainSyncEvent.getInt() => int j;
				
				// 0 = internal, 1 = midi sync
				if(i == 0) {
					//<<< "Moving to internal timer" >>>;
					mainS.exit();
					spork ~ main() @=> mainS;

				} else {
					//<<< "Moving to MIDI sync" >>>;
					mainS.exit();
					openMidiInDevice(j);
					spork ~ mainSync() @=> mainS;
				}
			}
		}
	}
	spork ~ oscMainSyncListener();
	
	
	while(true) {
		0.5::second => now;
	}
		